@@ -3,12 +3,11 @@ import argparse
 import os
 import warnings
 
+import coverage
 import pytest
-from coverage.misc import CoverageException
 
 from . import compat
 from . import embed
-from . import engine
 
 
 class CoverageError(Exception):
@@ -41,6 +40,21 @@ def validate_report(arg):
     return values
 
 
+def validate_fail_under(num_str):
+    try:
+        return int(num_str)
+    except ValueError:
+        return float(num_str)
+
+
+def validate_context(arg):
+    if coverage.version_info <= (5, 0):
+        raise argparse.ArgumentTypeError('Contexts are only supported with coverage.py >= 5.x')
+    if arg != "test":
+        raise argparse.ArgumentTypeError('--cov-context=test is the only supported value')
+    return arg
+
+
 class StoreReport(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
         report_type, file = values
@@ -73,13 +87,17 @@ def pytest_addoption(parser):
     group.addoption('--no-cov', action='store_true', default=False,
                     help='Disable coverage report completely (useful for debuggers). '
                          'Default: False')
-    group.addoption('--cov-fail-under', action='store', metavar='MIN', type=int,
+    group.addoption('--cov-fail-under', action='store', metavar='MIN',
+                    type=validate_fail_under,
                     help='Fail if the total coverage is less than MIN.')
     group.addoption('--cov-append', action='store_true', default=False,
                     help='Do not delete coverage but append to current. '
                          'Default: False')
     group.addoption('--cov-branch', action='store_true', default=None,
                     help='Enable branch coverage.')
+    group.addoption('--cov-context', action='store', metavar='CONTEXT',
+                    type=validate_context,
+                    help='Dynamic contexts to use. "test" for now.')
 
 
 def _prepare_cov_source(cov_source):
@@ -94,8 +112,18 @@ def _prepare_cov_source(cov_source):
 
 @pytest.mark.tryfirst
 def pytest_load_initial_conftests(early_config, parser, args):
+    options = early_config.known_args_namespace
+    no_cov = options.no_cov_should_warn = False
+    for arg in args:
+        arg = str(arg)
+        if arg == '--no-cov':
+            no_cov = True
+        elif arg.startswith('--cov') and no_cov:
+            options.no_cov_should_warn = True
+            break
+
     if early_config.known_args_namespace.cov_source:
-        plugin = CovPlugin(early_config.known_args_namespace, early_config.pluginmanager)
+        plugin = CovPlugin(options, early_config.pluginmanager)
         early_config.pluginmanager.register(plugin, '_cov')
 
 
@@ -104,10 +132,10 @@ class CovPlugin(object):
 
     Delegates all work to a particular implementation based on whether
     this test process is centralised, a distributed master or a
-    distributed slave.
+    distributed worker.
     """
 
-    def __init__(self, options, pluginmanager, start=True):
+    def __init__(self, options, pluginmanager, start=True, no_cov_should_warn=False):
         """Creates a coverage pytest plugin.
 
         We read the rc file that coverage uses to get the data file
@@ -122,6 +150,7 @@ class CovPlugin(object):
         self.cov_total = None
         self.failed = False
         self._started = False
+        self._start_path = None
         self._disabled = False
         self.options = options
 
@@ -138,12 +167,16 @@ class CovPlugin(object):
             self.options.cov_report = {}
         self.options.cov_source = _prepare_cov_source(self.options.cov_source)
 
+        # import engine lazily here to avoid importing
+        # it for unit tests that don't need it
+        from . import engine
+
         if is_dist and start:
             self.start(engine.DistMaster)
         elif start:
             self.start(engine.Central)
 
-        # slave is started in pytest hook
+        # worker is started in pytest hook
 
     def start(self, controller_cls, config=None, nodeid=None):
 
@@ -165,12 +198,13 @@ class CovPlugin(object):
         )
         self.cov_controller.start()
         self._started = True
+        self._start_path = os.getcwd()
         cov_config = self.cov_controller.cov.config
         if self.options.cov_fail_under is None and hasattr(cov_config, 'fail_under'):
             self.options.cov_fail_under = cov_config.fail_under
 
-    def _is_slave(self, session):
-        return hasattr(session.config, 'slaveinput')
+    def _is_worker(self, session):
+        return getattr(session.config, 'workerinput', None) is not None
 
     def pytest_sessionstart(self, session):
         """At session start determine our implementation and delegate to it."""
@@ -180,14 +214,22 @@ class CovPlugin(object):
             self._disabled = True
             return
 
+        # import engine lazily here to avoid importing
+        # it for unit tests that don't need it
+        from . import engine
+
         self.pid = os.getpid()
-        if self._is_slave(session):
-            nodeid = session.config.slaveinput.get('slaveid',
-                                                   getattr(session, 'nodeid'))
-            self.start(engine.DistSlave, session.config, nodeid)
+        if self._is_worker(session):
+            nodeid = (
+                session.config.workerinput.get('workerid', getattr(session, 'nodeid'))
+            )
+            self.start(engine.DistWorker, session.config, nodeid)
         elif not self._started:
             self.start(engine.Central)
 
+        if self.options.cov_context == 'test':
+            session.config.pluginmanager.register(TestContextPlugin(self.cov_controller.cov), '_cov_contexts')
+
     def pytest_configure_node(self, node):
         """Delegate to our implementation.
 
@@ -228,17 +270,19 @@ class CovPlugin(object):
         if self.cov_controller is not None:
             self.cov_controller.finish()
 
-        if not self._is_slave(session) and self._should_report():
+        if not self._is_worker(session) and self._should_report():
+
+            # import coverage lazily here to avoid importing
+            # it for unit tests that don't need it
+            from coverage.misc import CoverageException
+
             try:
                 self.cov_total = self.cov_controller.summary(self.cov_report)
             except CoverageException as exc:
                 message = 'Failed to generate report: %s\n' % exc
                 session.config.pluginmanager.getplugin("terminalreporter").write(
                     'WARNING: %s\n' % message, red=True, bold=True)
-                if pytest.__version__ >= '3.8':
-                    warnings.warn(pytest.PytestWarning(message))
-                else:
-                    session.config.warn(code='COV-2', message=message)
+                warnings.warn(pytest.PytestWarning(message))
                 self.cov_total = 0
             assert self.cov_total is not None, 'Test coverage should never be `None`'
             if self._failed_cov_total():
@@ -247,12 +291,10 @@ class CovPlugin(object):
 
     def pytest_terminal_summary(self, terminalreporter):
         if self._disabled:
-            message = 'Coverage disabled via --no-cov switch!'
-            terminalreporter.write('WARNING: %s\n' % message, red=True, bold=True)
-            if pytest.__version__ >= '3.8':
+            if self.options.no_cov_should_warn:
+                message = 'Coverage disabled via --no-cov switch!'
+                terminalreporter.write('WARNING: %s\n' % message, red=True, bold=True)
                 warnings.warn(pytest.PytestWarning(message))
-            else:
-                terminalreporter.config.warn(code='COV-1', message=message)
             return
         if self.cov_controller is None:
             return
@@ -264,20 +306,18 @@ class CovPlugin(object):
         terminalreporter.write('\n' + self.cov_report.getvalue() + '\n')
 
         if self.options.cov_fail_under is not None and self.options.cov_fail_under > 0:
-            if self.cov_total < self.options.cov_fail_under:
-                markup = {'red': True, 'bold': True}
-                message = (
-                    'FAIL Required test coverage of %d%% not '
-                    'reached. Total coverage: %.2f%%\n'
-                    % (self.options.cov_fail_under, self.cov_total)
-                )
-            else:
-                markup = {'green': True}
-                message = (
-                    'Required test coverage of %d%% '
-                    'reached. Total coverage: %.2f%%\n'
-                    % (self.options.cov_fail_under, self.cov_total)
+            failed = self.cov_total < self.options.cov_fail_under
+            markup = {'red': True, 'bold': True} if failed else {'green': True}
+            message = (
+                '{fail}Required test coverage of {required}% {reached}. '
+                'Total coverage: {actual:.2f}%\n'
+                .format(
+                    required=self.options.cov_fail_under,
+                    actual=self.cov_total,
+                    fail="FAIL " if failed else "",
+                    reached="not reached" if failed else "reached"
                 )
+            )
             terminalreporter.write(message, **markup)
 
     def pytest_runtest_setup(self, item):
@@ -300,8 +340,27 @@ class CovPlugin(object):
             yield
 
 
+class TestContextPlugin(object):
+    def __init__(self, cov):
+        self.cov = cov
+
+    def pytest_runtest_setup(self, item):
+        self.switch_context(item, 'setup')
+
+    def pytest_runtest_teardown(self, item):
+        self.switch_context(item, 'teardown')
+
+    def pytest_runtest_call(self, item):
+        self.switch_context(item, 'run')
+
+    def switch_context(self, item, when):
+        context = "{item.nodeid}|{when}".format(item=item, when=when)
+        self.cov.switch_context(context)
+
+
 @pytest.fixture
 def no_cover():
+    """A pytest fixture to disable coverage."""
     pass
 
 
